"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import cfnlint.helpers
from cfnlint.helpers import RESOURCE_SPECS
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class ValueRefGetAtt(CloudFormationLintRule):
    """Check if Resource Properties are correct"""

    id = "E3008"
    shortdesc = "Check values of properties for valid Refs and GetAtts"
    description = "Checks resource properties for Ref and GetAtt values"
    tags = ["resources", "ref", "getatt"]

    def initialize(self, cfn):
        """Initialize the rule"""
        for resource_type_spec in RESOURCE_SPECS.get(cfn.regions[0]).get(
            "ResourceTypes"
        ):
            self.resource_property_types.append(resource_type_spec)
        for property_type_spec in RESOURCE_SPECS.get(cfn.regions[0]).get(
            "PropertyTypes"
        ):
            self.resource_sub_property_types.append(property_type_spec)

    def is_value_a_list(self, path, property_name):
        """
        Determines if the value checked is a list or a value in a list
        We need to handle conditions in the path that could be nested, etc.
        ['Resources', 'LoadBalancer', 'Properties', 'Subnets', 'Fn::If', 2, 'Fn::If', 2]
        Numbers preceeded by a Fn::If should be removed and check repeated.
        """
        if path[-1] != property_name:
            # Property doesn't match the property name
            # Check if its a number and a condition
            if isinstance(path[-1], int) and path[-2] == "Fn::If":
                return self.is_value_a_list(path[:-2], property_name)

            return False

        return True

    def check_value_ref(self, value, path, **kwargs):
        """Check Ref"""
        matches = []
        cfn = kwargs.get("cfn")
        value_specs = kwargs.get("value_specs", {}).get("Ref")
        list_value_specs = kwargs.get("list_value_specs", {}).get("Ref")
        property_type = kwargs.get("property_type")
        property_name = kwargs.get("property_name")
        if (
            path[-1] == "Ref"
            and property_type == "List"
            and self.is_value_a_list(path[:-1], property_name)
        ):
            specs = list_value_specs
        else:
            specs = value_specs

        if not specs:
            # If no Ref's are specified, just skip
            # Opposite of GetAtt you will always have a Ref to a Parameter so if this is
            # None it just hasn't been defined and we can skip
            return matches

        if value in cfn.template.get("Parameters", {}):
            param = cfn.template.get("Parameters").get(value, {})
            parameter_type = param.get("Type")
            valid_parameter_types = []
            for parameter in specs.get("Parameters"):
                for param_type in (
                    RESOURCE_SPECS.get(cfn.regions[0])
                    .get("ParameterTypes")
                    .get(parameter)
                ):
                    valid_parameter_types.append(param_type)

            if not specs.get("Parameters"):
                message = 'Property "{0}" has no valid Refs to Parameters at {1}'
                matches.append(
                    RuleMatch(
                        path, message.format(property_name, "/".join(map(str, path)))
                    )
                )
            elif parameter_type not in valid_parameter_types:
                message = 'Property "{0}" can Ref to parameter of types [{1}] at {2}'
                matches.append(
                    RuleMatch(
                        path,
                        message.format(
                            property_name,
                            ", ".join(map(str, valid_parameter_types)),
                            "/".join(map(str, path)),
                        ),
                    )
                )
        if value in cfn.template.get("Resources", {}):
            resource = cfn.template.get("Resources").get(value, {})
            resource_type = resource.get("Type")
            if not specs.get("Resources"):
                message = 'Property "{0}" has no valid Refs to Resources at {1}'
                matches.append(
                    RuleMatch(
                        path, message.format(property_name, "/".join(map(str, path)))
                    )
                )
            elif resource_type not in specs.get("Resources"):
                message = 'Property "{0}" can Ref to resources of types [{1}] at {2}'
                matches.append(
                    RuleMatch(
                        path,
                        message.format(
                            property_name,
                            ", ".join(map(str, specs.get("Resources"))),
                            "/".join(map(str, path)),
                        ),
                    )
                )

        return matches

    def check_value_getatt(self, value, path, **kwargs):
        """Check GetAtt"""
        matches = []
        cfn = kwargs.get("cfn")
        value_specs = kwargs.get("value_specs", {}).get("GetAtt")
        list_value_specs = kwargs.get("list_value_specs", {}).get("GetAtt")
        property_type = kwargs.get("property_type")
        property_name = kwargs.get("property_name")
        # You can sometimes get a list or a string with . in it
        if isinstance(value, list):
            resource_name = value[0]
            if len(value[1:]) == 1:
                resource_attribute = value[1].split(".")
            else:
                resource_attribute = value[1:]
        elif isinstance(value, str):
            resource_name = value.split(".")[0]
            resource_attribute = value.split(".")[1:]
        is_value_a_list = self.is_value_a_list(path[:-1], property_name)
        if path[-1] == "Fn::GetAtt" and property_type == "List" and is_value_a_list:
            specs = list_value_specs
        else:
            specs = value_specs

        resource_type = (
            cfn.template.get("Resources", {}).get(resource_name, {}).get("Type")
        )

        if cfnlint.helpers.is_custom_resource(resource_type):
            #  A custom resource voids the spec.  Move on
            return matches

        if (
            resource_type
            in [
                "AWS::CloudFormation::Stack",
                "AWS::ServiceCatalog::CloudFormationProvisionedProduct",
            ]
            and resource_attribute[0] == "Outputs"
        ):
            # Nested Stack Outputs
            # if its a string type we are good and return matches
            # if its a list its a failure as Outputs can only be strings
            if is_value_a_list and property_type == "List":
                message = (
                    "CloudFormation stack outputs need to be strings not lists at {0}"
                )
                matches.append(
                    RuleMatch(path, message.format("/".join(map(str, path))))
                )

            return matches

        if specs is None:
            # GetAtt specs aren't specified skip
            return matches
        if not specs:
            # GetAtt is specified but empty so there are no valid options
            message = 'Property "{0}" has no valid Fn::GetAtt options at {1}'
            matches.append(
                RuleMatch(path, message.format(property_name, "/".join(map(str, path))))
            )
            return matches

        if resource_type not in specs:
            message = (
                'Property "{0}" can Fn::GetAtt to a resource of types [{1}] at {2}'
            )
            matches.append(
                RuleMatch(
                    path,
                    message.format(
                        property_name,
                        ", ".join(map(str, specs)),
                        "/".join(map(str, path)),
                    ),
                )
            )
        elif isinstance(specs[resource_type], list):
            found = False
            for allowed_att in specs[resource_type]:
                if ".".join(map(str, resource_attribute)) == allowed_att:
                    found = True
            if not found:
                message = (
                    'Property "{0}" can Fn::GetAtt to a resource attribute "{1}" at {2}'
                )
                matches.append(
                    RuleMatch(
                        path,
                        message.format(
                            property_name,
                            specs[resource_type],
                            "/".join(map(str, path)),
                        ),
                    )
                )
        elif ".".join(map(str, resource_attribute)) != specs[resource_type]:
            message = (
                'Property "{0}" can Fn::GetAtt to a resource attribute "{1}" at {2}'
            )
            matches.append(
                RuleMatch(
                    path,
                    message.format(
                        property_name, specs[resource_type], "/".join(map(str, path))
                    ),
                )
            )

        return matches

    def _get_value_specs(self, value_type, region):
        value_specs = RESOURCE_SPECS.get(region).get("ValueTypes").get(value_type, {})
        if value_specs == "CACHED":
            value_specs = (
                RESOURCE_SPECS.get("us-east-1").get("ValueTypes").get(value_type, {})
            )
        return value_specs

    def check(self, cfn, properties, value_specs, property_specs, path):
        """Check itself"""
        matches = []
        for p_value, p_path in properties.items_safe(path[:]):
            for prop in p_value:
                if prop in value_specs:
                    value = value_specs.get(prop).get("Value", {})
                    if value:
                        value_type = value.get("ValueType", "")
                        list_value_type = value.get("ListValueType", "")
                        property_type = (
                            property_specs.get("Properties").get(prop).get("Type")
                        )
                        value_type_specs = self._get_value_specs(
                            value_type, cfn.regions[0]
                        )
                        list_value_specs = self._get_value_specs(
                            list_value_type, cfn.regions[0]
                        )
                        matches.extend(
                            cfn.check_value(
                                p_value,
                                prop,
                                p_path,
                                check_ref=self.check_value_ref,
                                check_get_att=self.check_value_getatt,
                                value_specs=value_type_specs,
                                list_value_specs=list_value_specs,
                                cfn=cfn,
                                property_type=property_type,
                                property_name=prop,
                            )
                        )

        return matches

    def match_resource_sub_properties(self, properties, property_type, path, cfn):
        """Match for sub properties"""
        matches = []

        specs = (
            RESOURCE_SPECS.get(cfn.regions[0])
            .get("PropertyTypes")
            .get(property_type, {})
            .get("Properties", {})
        )
        property_specs = (
            RESOURCE_SPECS.get(cfn.regions[0]).get("PropertyTypes").get(property_type)
        )
        matches.extend(self.check(cfn, properties, specs, property_specs, path))

        return matches

    def match_resource_properties(self, properties, resource_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        specs = (
            RESOURCE_SPECS.get(cfn.regions[0])
            .get("ResourceTypes")
            .get(resource_type, {})
            .get("Properties", {})
        )
        resource_specs = (
            RESOURCE_SPECS.get(cfn.regions[0]).get("ResourceTypes").get(resource_type)
        )
        matches.extend(self.check(cfn, properties, specs, resource_specs, path))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
# pylint: disable=cyclic-import
import cfnlint.rules

OPERATOR = ["EQUALS", "NOT_EQUALS", "==", "!=", "IN", "NOT_IN", ">=", "<="]


def CreateCustomRule(
    rule_id, resourceType, prop, value, error_message, description, shortdesc, rule_func
):
    class CustomRule(cfnlint.rules.CloudFormationLintRule):
        def __init__(
            self,
            rule_id,
            resourceType,
            prop,
            value,
            error_message,
            description,
            shortdesc,
            rule_func,
        ):
            super().__init__()
            self.resource_property_types.append(resourceType)
            self.id = rule_id
            self.property_chain = prop.split(".")
            self.property_value = value
            self.error_message = error_message
            self.description = description
            self.shortdesc = shortdesc
            self.rule_func = rule_func

        def _remaining_inset_properties(self, property_chain):
            if len(property_chain) > 1:
                return property_chain[1:]

            return []

        def _check_value(self, value, path, property_chain, cfn):
            matches = []
            if property_chain:
                new_property_chain = self._remaining_inset_properties(property_chain)
                matches.extend(
                    cfn.check_value(
                        value,
                        property_chain[0],
                        path,
                        check_value=self._check_value,
                        property_chain=new_property_chain,
                        cfn=cfn,
                    )
                )
                return matches
            if value is not None:
                matches.extend(self.rule_func(value, self.property_value, path))
            return matches

        def match_resource_properties(self, properties, _, path, cfn):
            new_property_chain = self._remaining_inset_properties(self.property_chain)
            return cfn.check_value(
                properties,
                self.property_chain[0],
                path,
                check_value=self._check_value,
                property_chain=new_property_chain,
                cfn=cfn,
            )

    return CustomRule(
        rule_id,
        resourceType,
        prop,
        value,
        error_message,
        description,
        shortdesc,
        rule_func,
    )


def CreateEqualsRule(rule_id, resourceType, prop, value, error_message):
    def rule_func(value, expected_value, path):
        matches = []
        if str(value).strip().lower() != str(expected_value).strip().lower():
            matches.append(
                cfnlint.rules.RuleMatch(
                    path, error_message or "Must equal check failed"
                )
            )

        return matches

    return CreateCustomRule(
        rule_id,
        resourceType,
        prop,
        value,
        error_message,
        shortdesc="Custom rule to check for equal values",
        description="Created from the custom rules parameter. This rule will check if a property value is equal to the specified value.",
        rule_func=rule_func,
    )


def CreateNotEqualsRule(rule_id, resourceType, prop, value, error_message):
    def rule_func(value, expected_values, path):
        matches = []
        if str(value).strip().lower() == str(expected_values).strip().lower():
            matches.append(
                cfnlint.rules.RuleMatch(
                    path, error_message or "Must not equal check failed"
                )
            )

        return matches

    return CreateCustomRule(
        rule_id,
        resourceType,
        prop,
        value,
        error_message,
        shortdesc="Custom rule to check for not equal values",
        description="Created from the custom rules parameter. This rule will check if a property value is NOT equal to the specified value.",
        rule_func=rule_func,
    )


def CreateGreaterRule(rule_id, resourceType, prop, value, error_message):
    def rule_func(value, expected_value, path):
        matches = []
        if checkInt(str(value).strip()) and checkInt(str(expected_value).strip()):
            if int(str(value).strip()) < int(str(expected_value).strip()):
                matches.append(
                    cfnlint.rules.RuleMatch(
                        path, error_message or "Greater than check failed"
                    )
                )
        else:
            matches.append(
                cfnlint.rules.RuleMatch(
                    path, error_message or "Given values are not numeric"
                )
            )

        return matches

    return CreateCustomRule(
        rule_id,
        resourceType,
        prop,
        value,
        error_message,
        shortdesc="Custom rule to check for if a value is greater than the specified value",
        description="Created from the custom rules parameter. This rule will check if a property value is greater than the specified value.",
        rule_func=rule_func,
    )


def CreateLesserRule(rule_id, resourceType, prop, value, error_message):
    def rule_func(value, expected_value, path):
        matches = []
        if checkInt(str(value).strip()) and checkInt(str(expected_value).strip()):
            if int(str(value).strip()) > int(str(expected_value).strip()):
                matches.append(
                    cfnlint.rules.RuleMatch(
                        path, error_message or "Lesser than check failed"
                    )
                )
        else:
            matches.append(
                cfnlint.rules.RuleMatch(
                    path, error_message or "Given values are not numeric"
                )
            )

        return matches

    return CreateCustomRule(
        rule_id,
        resourceType,
        prop,
        value,
        error_message,
        shortdesc="Custom rule to check for if a value is lesser than the specified value",
        description="Created from the custom rules parameter. This rule will check if a property value is lesser than the specified value.",
        rule_func=rule_func,
    )


def CreateInSetRule(rule_id, resourceType, prop, value, error_message):
    def rule_func(value, expected_values, path):
        matches = []
        if value not in expected_values:
            matches.append(
                cfnlint.rules.RuleMatch(path, error_message or "In set check failed")
            )

        return matches

    return CreateCustomRule(
        rule_id,
        resourceType,
        prop,
        value,
        error_message,
        shortdesc="Custom rule to check for if a value exists in a list of specified values",
        description="Created from the custom rules parameter. This rule will check if a property value exists inside a list of specified values.",
        rule_func=rule_func,
    )


def CreateNotInSetRule(rule_id, resourceType, prop, value, error_message):
    def rule_func(value, expected_values, path):
        matches = []
        if value in expected_values:
            matches.append(
                cfnlint.rules.RuleMatch(
                    path, error_message or "Not in set check failed"
                )
            )

        return matches

    return CreateCustomRule(
        rule_id,
        resourceType,
        prop,
        value,
        error_message,
        shortdesc="Custom rule to check for if a value does not exist in a list of specified values",
        description="Created from the custom rules parameter. This rule will check if a property value does not exist inside a list of specified values.",
        rule_func=rule_func,
    )


def CreateInvalidRule(rule_id, operator):
    class InvalidRule(cfnlint.rules.CloudFormationLintRule):
        def __init__(self, rule_id, operator):
            super().__init__()
            self.id = rule_id
            self.operator = operator
            self.description = "Created from the custom rule parameter. This rule is the result of an invalid configuration of a custom rule."
            self.shortdesc = "Invalid custom rule configuration"

        def match(self, _):
            message = '"{0}" not in supported operators: [{1}]'
            return [
                cfnlint.rules.RuleMatch(
                    [], message.format(str(self.operator), ", ".join(OPERATOR))
                )
            ]

    return InvalidRule(rule_id, operator)


def checkInt(i):
    """Python 2.7 Compatibility - There is no isnumeric() method"""
    try:
        int(i)
        return True
    except ValueError:
        return False



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import fileinput
import json
import logging
import sys
from json.decoder import (  # type: ignore
    BACKSLASH,
    STRINGCHUNK,
    WHITESPACE,
    WHITESPACE_STR,
)
from json.scanner import NUMBER_RE

import cfnlint
from cfnlint.decode.node import dict_node, list_node, str_node, sub_node

LOGGER = logging.getLogger(__name__)


class DuplicateError(Exception):
    """
    Error thrown when the template contains duplicates
    """

    def __init__(self, message, key):
        super().__init__(message)
        self.duplicates = []
        self.append(key)

    def append(self, key):
        self.duplicates.append(key)


def check_duplicates(ordered_pairs, beg_mark, end_mark):
    """
    Check for duplicate keys on the current level, this is not desirable
    because a dict does not support this. It overwrites it with the last
    occurrence, which can give unexpected results
    """
    err = None
    mapping = dict_node({}, beg_mark, end_mark)
    for key, value in ordered_pairs:
        for m_key in mapping:
            if m_key == key:
                if err is None:
                    err = DuplicateError(f'"{key}"', key)
                    err.append(m_key)
                else:
                    err.append(key)
        if key not in mapping:
            mapping[key] = value

    if err is not None:
        raise err

    if len(mapping) == 1:
        if "Fn::Sub" in mapping:
            return sub_node(ordered_pairs, beg_mark, end_mark)
    return mapping


class JSONDecodeError(ValueError):
    """Subclass of ValueError with the following additional properties:
    msg: The unformatted error message
    doc: The JSON document being parsed
    pos: The start index of doc where parsing failed
    lineno: The line corresponding to pos
    colno: The column corresponding to pos
    """

    # Note that this exception is used from _json

    def __init__(self, doc, pos, errors):
        if isinstance(errors, cfnlint.rules.Match):
            errors = [errors]

        errmsg = f"{errors[0].message}: line {errors[0].linenumber} column {errors[0].linenumber} (char {pos})"
        ValueError.__init__(self, errmsg)
        self.msg = errors[0].message
        self.doc = doc
        self.pos = pos
        self.lineno = errors[0].linenumber
        self.colno = errors[0].linenumber
        self.matches = errors

    def __reduce__(self):
        return self.__class__, (self.msg, self.doc, self.pos)


def build_match(message, doc, pos, key=" "):
    lineno = doc.count("\n", 0, pos) + 1
    colno = pos - doc.rfind("\n", 0, pos)

    return cfnlint.rules.Match(
        lineno,
        colno + 1,
        lineno,
        colno + 1 + len(key),
        "",
        cfnlint.rules.ParseError(),
        message=message,
    )


def build_match_from_key(message, key):
    return cfnlint.rules.Match(
        key.start_mark.line,
        key.start_mark.column + 1,
        key.end_mark.line,
        key.end_mark.column + 1 + len(key),
        "",
        cfnlint.rules.ParseError(),
        message=message,
    )


class Mark:
    """Mark of line and column"""

    line = 1
    column = 1

    def __init__(self, line, column):
        self.line = line
        self.column = column


# pylint: disable=W0102
# Exception based on builtin Python Function
def py_scanstring(s, end, strict=True, _b=BACKSLASH, _m=STRINGCHUNK.match):
    """Scan the string s for a JSON string. End is the index of the
    character in s after the quote that started the JSON string.
    Unescapes all valid JSON string escape sequences and raises ValueError
    on attempt to decode an invalid string. If strict is False then literal
    control characters are allowed in the string.
    Returns a tuple of the decoded string and the index of the character in s
    after the end quote."""
    chunks = []
    _append = chunks.append
    begin = end - 1
    while 1:
        chunk = _m(s, end)
        if chunk is None:
            raise JSONDecodeError(
                doc=s,
                pos=begin,
                errors=[
                    build_match(
                        message="Unterminated string starting at",
                        doc=s,
                        pos=begin,
                    ),
                ],
            )
        end = chunk.end()
        content, terminator = chunk.groups()
        # Content is contains zero or more unescaped string characters
        if content:
            _append(content)
        # Terminator is the end of string, a literal control character,
        # or a backslash denoting that an escape sequence follows
        if terminator == '"':
            break
        if terminator != "\\":
            if strict:
                msg = f"Invalid control character {terminator!r} at"
                raise JSONDecodeError(
                    doc=s,
                    pos=end,
                    errors=[
                        build_match(
                            message=msg,
                            doc=s,
                            pos=end,
                        ),
                    ],
                )
            _append(terminator)
            continue
        try:
            esc = s[end]
        except IndexError as exc:
            raise JSONDecodeError(
                doc=s,
                pos=begin,
                errors=[
                    build_match(
                        message="Unterminated string starting at",
                        doc=s,
                        pos=begin,
                    ),
                ],
            ) from exc
        # If not a unicode escape sequence, must be in the lookup table
        if esc != "u":
            try:
                char = _b[esc]
            except KeyError as exc:
                msg = f"Invalid \\escape: {esc!r}"
                raise JSONDecodeError(
                    doc=s,
                    pos=end,
                    errors=[
                        build_match(
                            message=msg,
                            doc=s,
                            pos=end,
                        ),
                    ],
                ) from exc
            end += 1
        else:
            uni = _decode_uXXXX(s, end)
            end += 5
            if 0xD800 <= uni <= 0xDBFF and s[end : end + 2] == "\\u":
                uni2 = _decode_uXXXX(s, end + 1)
                if 0xDC00 <= uni2 <= 0xDFFF:
                    uni = 0x10000 + (((uni - 0xD800) << 10) | (uni2 - 0xDC00))
                    end += 6
            char = chr(uni)
        _append(char)
    return "".join(chunks), end


def _decode_uXXXX(s, pos):
    esc = s[pos + 1 : pos + 5]
    if len(esc) == 4 and esc[1] not in "xX":
        try:
            return int(esc, 16)
        except ValueError:
            pass
    msg = "Invalid \\uXXXX escape"
    raise JSONDecodeError(
        doc=s,
        pos=pos,
        errors=[
            build_match(
                message=msg,
                doc=s,
                pos=pos,
            ),
        ],
    )


def py_make_scanner(context):
    """
    Make python based scanner
    For this use case we will not use the C based scanner
    """
    parse_object = context.parse_object
    parse_array = context.parse_array
    parse_string = context.parse_string
    match_number = NUMBER_RE.match
    strict = context.strict
    parse_float = context.parse_float
    parse_int = context.parse_int
    parse_constant = context.parse_constant
    object_hook = context.object_hook
    object_pairs_hook = context.object_pairs_hook
    memo = context.memo

    # pylint: disable=R0911
    # Based on Python standard function
    def _scan_once(string, idx):
        """Scan once internal function"""
        try:
            nextchar = string[idx]
        except IndexError as exc:
            raise StopIteration(idx) from exc

        if nextchar == '"':
            return parse_string(string, idx + 1, strict)
        if nextchar == "{":
            return parse_object(
                (string, idx + 1),
                strict,
                scan_once,
                object_hook,
                object_pairs_hook,
                memo,
            )
        if nextchar == "[":
            return parse_array((string, idx + 1), _scan_once)
        if nextchar == "n" and string[idx : idx + 4] == "null":
            return None, idx + 4
        if nextchar == "t" and string[idx : idx + 4] == "true":
            return True, idx + 4
        if nextchar == "f" and string[idx : idx + 5] == "false":
            return False, idx + 5

        m = match_number(string, idx)
        if m is not None:
            integer, frac, exp = m.groups()
            if frac or exp:
                res = parse_float(integer + (frac or "") + (exp or ""))
            else:
                res = parse_int(integer)
            return res, m.end()
        if nextchar == "N" and string[idx : idx + 3] == "NaN":
            return parse_constant("NaN"), idx + 3
        if nextchar == "I" and string[idx : idx + 8] == "Infinity":
            return parse_constant("Infinity"), idx + 8
        if nextchar == "-" and string[idx : idx + 9] == "-Infinity":
            return parse_constant("-Infinity"), idx + 9

        raise StopIteration(idx)

    def scan_once(string, idx):
        """Scan Once"""
        try:
            return _scan_once(string, idx)
        finally:
            memo.clear()

    return _scan_once


def find_indexes(s, ch="\n"):
    """Finds all instances of given char and returns list of indexes"""
    return [i for i, ltr in enumerate(s) if ltr == ch]


def count_occurrences(arr, key):
    """Binary search indexes to replace str.count"""
    n = len(arr)
    left = 0
    right = n - 1
    count = 0

    while left <= right:
        mid = int((right + left) / 2)

        if arr[mid] <= key:
            count = mid + 1
            left = mid + 1
        else:
            right = mid - 1
    return count


def largest_less_than(indexes, line_num, pos):
    """Replacement func for python str.rfind using indexes"""
    return indexes[line_num - 1] if indexes and count_occurrences(indexes, pos) else -1


def get_beg_end_mark(start, end, indexes):
    """Get the Start and End Mark"""
    beg_lineno = count_occurrences(indexes, start)
    beg_colno = start - largest_less_than(indexes, beg_lineno, start)
    beg_mark = Mark(beg_lineno, beg_colno)

    offset = 1 if len(indexes) > 1 else 0
    end_lineno = count_occurrences(indexes, end) - offset
    end_colno = end - largest_less_than(indexes, end_lineno, end)
    end_mark = Mark(end_lineno, end_colno)

    return beg_mark, end_mark


def load(filename):
    """
    Load the given JSON file
    """

    content = ""

    if not sys.stdin.isatty():
        filename = "-" if filename is None else filename
        if sys.version_info.major <= 3 and sys.version_info.minor <= 9:
            for line in fileinput.input(files=filename):
                content = content + line
        else:
            for line in fileinput.input(  # pylint: disable=unexpected-keyword-arg
                files=filename, encoding="utf-8"
            ):
                content = content + line
    else:
        with open(filename, encoding="utf-8") as fp:
            content = fp.read()

    return json.loads(content, cls=CfnJSONDecoder)


def loads(json_string):
    """
    Load the given JSON string
    """
    return json.loads(json_string, cls=CfnJSONDecoder)


class CfnJSONDecoder(json.JSONDecoder):
    """
    Converts a json string, where datetime and timedelta objects were converted
    into strings using the DateTimeAwareJSONEncoder, into a python object.
    """

    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, *args, **kwargs)
        self.parse_object = self.cfn_json_object
        self.parse_array = self.JSONArray
        self.parse_string = py_scanstring
        self.memo = {}
        self.object_pairs_hook = check_duplicates
        self.scan_once = py_make_scanner(self)
        self.newline_indexes = []

    def decode(self, s, _w=WHITESPACE.match):
        """Overridden to retrieve indexes"""
        self.newline_indexes = find_indexes(s)
        obj = super().decode(s, _w)
        return obj

    def JSONArray(self, s_and_end, scan_once, **kwargs):
        """Convert JSON array to be a list_node object"""
        values, end = json.decoder.JSONArray(s_and_end, scan_once, **kwargs)
        start = s_and_end[1]
        beg_mark, end_mark = get_beg_end_mark(start, end, self.newline_indexes)
        return list_node(values, beg_mark, end_mark), end

    def cfn_json_object(
        self,
        s_and_end,
        strict,
        scan_once,
        object_hook,
        object_pairs_hook,
        memo=None,
        _w=WHITESPACE.match,
        _ws=WHITESPACE_STR,
    ):
        """Custom Cfn JSON Object to store keys with start and end times"""
        s, end = s_and_end
        orginal_end = end
        pairs = []
        pairs_append = pairs.append
        # Backwards compatibility
        if memo is None:
            memo = {}
        memo_get = memo.setdefault
        # Use a slice to prevent IndexError from being raised, the following
        # check will raise a more specific ValueError if the string is empty
        nextchar = s[end : end + 1]
        # Normally we expect nextchar == '"'
        if nextchar != '"':
            if nextchar in _ws:
                end = _w(s, end).end()
                nextchar = s[end : end + 1]
            # Trivial empty object
            if nextchar == "}":
                if object_pairs_hook is not None:
                    try:
                        beg_mark, end_mark = get_beg_end_mark(
                            orginal_end, end + 1, self.newline_indexes
                        )
                        result = object_pairs_hook(pairs, beg_mark, end_mark)
                        return result, end + 1
                    except DuplicateError as err:
                        errs = []
                        for m in err.duplicates:
                            errs.append(
                                build_match_from_key(
                                    message=f"Duplicate found {err}",
                                    key=m,
                                )
                            )
                        raise JSONDecodeError(
                            doc=s,
                            pos=end,
                            errors=errs,
                        ) from err
                pairs = {}
                if object_hook is not None:
                    beg_mark, end_mark = get_beg_end_mark(
                        orginal_end, end + 1, self.newline_indexes
                    )
                    pairs = object_hook(pairs, beg_mark, end_mark)
                return pairs, end + 1

            if nextchar != '"':
                raise JSONDecodeError(
                    doc=s,
                    pos=end,
                    errors=[
                        build_match(
                            message="Expecting property name enclosed in double quotes",
                            doc=s,
                            pos=end,
                        ),
                    ],
                )
        end += 1
        while True:
            begin = end - 1
            key, end = py_scanstring(s, end, strict)
            # print(lineno, colno, obj)
            # print(key, lineno, colno)
            key = memo_get(key, key)
            # To skip some function call overhead we optimize the fast paths where
            # the JSON key separator is ": " or just ":".
            if s[end : end + 1] != ":":
                end = _w(s, end).end()
                if s[end : end + 1] != ":":
                    raise JSONDecodeError(
                        doc=s,
                        pos=end,
                        errors=[
                            build_match(
                                message="Expecting ':' delimiter",
                                doc=s,
                                pos=end,
                            ),
                        ],
                    )
            end += 1

            try:
                if s[end] in _ws:
                    end += 1
                    if s[end] in _ws:
                        end = _w(s, end + 1).end()
            except IndexError:
                pass

            beg_mark, end_mark = get_beg_end_mark(
                begin, begin + len(key), self.newline_indexes
            )
            try:
                value, end = scan_once(s, end)
            except StopIteration as err:
                raise JSONDecodeError(
                    doc=s,
                    pos=str(err),
                    errors=[
                        build_match(
                            message="Expecting value",
                            doc=s,
                            pos=str(err),
                        ),
                    ],
                ) from err
            key_str = str_node(key, beg_mark, end_mark)
            pairs_append((key_str, value))
            try:
                nextchar = s[end]
                if nextchar in _ws:
                    end = _w(s, end + 1).end()
                    nextchar = s[end]
            except IndexError:
                nextchar = ""
            end += 1

            if nextchar == "}":
                break
            if nextchar != ",":
                raise JSONDecodeError(
                    doc=s,
                    pos=end - 1,
                    errors=[
                        build_match(
                            message="Expecting ',' delimiter",
                            doc=s,
                            pos=end - 1,
                        ),
                    ],
                )
            end = _w(s, end).end()
            nextchar = s[end : end + 1]
            end += 1
            if nextchar != '"':
                raise JSONDecodeError(
                    doc=s,
                    pos=end - 1,
                    errors=[
                        build_match(
                            message="Expecting property name enclosed in double quotes",
                            doc=s,
                            pos=end - 1,
                        ),
                    ],
                )
        if object_pairs_hook is not None:
            try:
                beg_mark, end_mark = get_beg_end_mark(
                    orginal_end, end, self.newline_indexes
                )
                result = object_pairs_hook(pairs, beg_mark, end_mark)
            except DuplicateError as err:
                errs = []
                for m in err.duplicates:
                    errs.append(
                        build_match_from_key(
                            message=f"Duplicate found {err}",
                            key=m,
                        )
                    )
                raise JSONDecodeError(
                    doc=s,
                    pos=end,
                    errors=errs,
                ) from err
            return result, end

        pairs = dict(pairs)
        if object_hook is not None:
            beg_mark, end_mark = get_beg_end_mark(
                orginal_end, end, self.newline_indexes
            )
            pairs = object_hook(pairs, beg_mark, end_mark)
        return pairs, end



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.helpers import FUNCTIONS_SINGLE
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Configuration(CloudFormationLintRule):
    """Check if Outputs are configured correctly"""

    id = "E6001"
    shortdesc = "Outputs have appropriate properties"
    description = "Making sure the outputs are properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/outputs-section-structure.html"
    tags = ["outputs"]

    valid_keys = ["Value", "Export", "Description", "Condition"]

    # Map can be singular or multiple for this case we are going to skip
    valid_funcs = FUNCTIONS_SINGLE + ["Fn::FindInMap", "Fn::Transform"]

    def check_func(self, value, path):
        """Check that a value is using the correct functions"""
        matches = []
        if isinstance(value, dict):
            if len(value) == 1:
                for k in value.keys():
                    if k not in self.valid_funcs:
                        message = "{0} must use one of the functions {1}"
                        matches.append(
                            RuleMatch(
                                path, message.format("/".join(path), self.valid_funcs)
                            )
                        )
            else:
                message = "{0} must use one of the functions {1}"
                matches.append(
                    RuleMatch(path, message.format("/".join(path), self.valid_funcs))
                )
        elif isinstance(value, (list)):
            message = "{0} must be a string or one of the functions {1}"
            matches.append(
                RuleMatch(path, message.format("/".join(path), self.valid_funcs))
            )

        return matches

    def check_export(self, value, path):
        """Check export structure"""
        matches = []
        if isinstance(value, dict):
            if len(value) == 1:
                for k, v in value.items():
                    if k != "Name":
                        message = '{0} must be a an object of one with key "Name"'
                        matches.append(RuleMatch(path, message.format("/".join(path))))
                    else:
                        matches.extend(self.check_func(v, path[:] + ["Name"]))
            else:
                message = '{0} must be a an object of one with key "Name"'
                matches.append(RuleMatch(path, message.format("/".join(path))))
        else:
            message = '{0} must be a an object of one with key "Name"'
            matches.append(RuleMatch(path, message.format("/".join(path))))

        return matches

    def match(self, cfn):
        matches = []

        outputs = cfn.template.get("Outputs", {})
        if outputs:
            if isinstance(outputs, dict):
                for output_name, output_value in outputs.items():
                    if not isinstance(output_value, dict):
                        message = "Output {0} is not an object"
                        matches.append(
                            RuleMatch(
                                ["Outputs", output_name], message.format(output_name)
                            )
                        )
                    else:
                        for propname, propvalue in output_value.items():
                            if propname not in self.valid_keys:
                                message = "Output {0} has invalid property {1}"
                                matches.append(
                                    RuleMatch(
                                        ["Outputs", output_name, propname],
                                        message.format(output_name, propname),
                                    )
                                )
                            else:
                                if propvalue is None:
                                    message = (
                                        "Output {0} has property {1} has invalid type"
                                    )
                                    matches.append(
                                        RuleMatch(
                                            ["Outputs", output_name, propname],
                                            message.format(output_name, propname),
                                        )
                                    )
                        value = output_value.get("Value")
                        if value:
                            matches.extend(
                                self.check_func(
                                    value, ["Outputs", output_name, "Value"]
                                )
                            )
                        export = output_value.get("Export")
                        if export:
                            matches.extend(
                                self.check_export(
                                    export, ["Outputs", output_name, "Export"]
                                )
                            )
            else:
                matches.append(
                    RuleMatch(["Outputs"], "Outputs do not follow correct format.")
                )

        return matches
